#ifndef __FIFOBINSTR_HPP
#define __FIFOBINSTR_HPP


#include <sys/types.h>
#include <sys/stat.h>
//#include <sys/sem.h>
//#include <sys/ipc.h>
//#include <sys/msg.h>
#include <string.h>
//#include <sys/shm.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <unistd.h>

//#define TEMP_FAILURE_RETRY(co) co

#define __USE_GNU
#include <unistd.h>
//#include <stdlib.h>

//__extension__
# define TEMP_FAILURE_RETRY(expression) \
  (							      \
    ({ long int __result;						      \
       do __result = (long int) (expression);				      \
       while (__result == -1L && errno == EINTR);			      \
       __result; }))



extern "C" void exit(int);


#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

//#include <hashtbl.hpp>
//#include <postlib.hpp>

#include "stm.hpp"

namespace stm {

class binstream  {
	public:
		virtual dword write(const void *co, dword size) =0;
		virtual dword read (void *co, dword size) =0;
		virtual dword flush() =0;

		virtual ~binstream() {}

};


class filebinstream : public  binstream {
	protected:
		int f;
		
	public:
		filebinstream(const char *fname) { f = checkerror(open(fname, O_RDWR)); }
		filebinstream(int fd) : f(fd) {  }  // takes ownership
		
		~filebinstream() { if (f != -1) close(f); }
		
		
		virtual dword write(const void *co, dword size)
			{ return checkerror(TEMP_FAILURE_RETRY(::write(f, co, size))); }
		virtual dword read (void *co, dword size)
			{ return checkerror(TEMP_FAILURE_RETRY(::read(f, co, size))); }
		virtual dword flush() { return 0; }
		
		
		int fd() { return f; }
		
		void putoff() { f = -1; }  // leaves the fd
		

};

class pipebinstr : public binstream {
	protected:
    	int reader;
        int writer;
	public:
		pipebinstr() { if (pipe(&reader) == -1) throw std_error(); }
		~pipebinstr() { close(reader); close(writer); }

        int readpipe() { return reader; }
		int writepipe() { return writer; }

		virtual dword write(const void *co, dword size)
			{ return ::write(writer, co, size); }
		virtual dword read (void *co, dword size)
			{ return ::read(reader, co, size); }
		virtual dword flush() { return 0; }

};


class npipebinstr : public binstream {
	protected:
    	int reader;
        int writer;
	public:
		npipebinstr() {}
		~npipebinstr() { close(reader); close(writer); }

        int readpipe() { return reader; }
		int writepipe() { return writer; }

		virtual dword write(const void *co, dword size)
			{ return TEMP_FAILURE_RETRY(::write(writer, co, size)); }
		virtual dword read (void *co, dword size)
			{ return TEMP_FAILURE_RETRY(::read(reader, co, size)); }
		virtual dword flush() { return 0; }

};



class socketbinstr : public binstream {
	protected:
    	int sock;

	public:
    	socketbinstr() {
	        sock = socket(PF_INET, SOCK_STREAM, 0);
        }
    	socketbinstr(int type) {
	        sock = socket(PF_INET, type, 0);
        }

        ~socketbinstr() {
        	close(sock);
        }
		virtual dword write(const void *co, dword size)
			{ return checkerror(TEMP_FAILURE_RETRY(::write(sock, co, size))); }
		virtual dword read (void *co, dword size)
			{ return checkerror(TEMP_FAILURE_RETRY(::read(sock, co, size))); }
		virtual dword flush() { return 0; }


};


class srvsocketbstr : public socketbinstr {
	private:
		int actsock;
		sockaddr_in addr;
	public:
		void create(word port) {
			sockaddr_in addr;
			inet_aton(INADDR_ANY, &addr.sin_addr);
			addr.sin_family = PF_INET;
			addr.sin_port = port;
			bind(sock, (sockaddr *)&addr, sizeof(sockaddr_in));
			listen(sock, 10);
		}



		srvsocketbstr(word port) {
		}

		virtual dword read (void *co, dword size) {
/*        	int r;
			do {
				if (r == -1)  { // connection broken, get new one
					close(actsock);
					int size = sizeof(sockaddr_in);
					actsock = accept(sock, (sockaddr *)&addr, &size);
				}
			} while (r == -1) */
			return ::read(sock, co, size);
		}
		virtual dword write(const void *co, dword size) {
			return TEMP_FAILURE_RETRY(::write(actsock, co, size));
		}
		virtual dword flush() { return 0; }
};

class clntsocketbstr : public socketbinstr {
	public:

		void create(word port, const char *haddr) {
			sockaddr_in addr;
			hostent *h;
			h = gethostbyname(haddr);

			if (h->h_addr_list[0])
				addr.sin_addr = *(in_addr *)h->h_addr_list[0];
			else
				throw std::runtime_error("No such host");

			addr.sin_family = PF_INET;
			addr.sin_port = port;

			connect(sock, (sockaddr *)&addr, sizeof(sockaddr_in));
	   }



		clntsocketbstr(word port, const char *haddr, int type = SOCK_STREAM) : socketbinstr(type)  {
			create(port, haddr);
		}

		clntsocketbstr(word port, in_addr haddr) {
			sockaddr_in addr;
			//inet_aton(haddr, &addr.sin_addr);
			addr.sin_addr = haddr;
			addr.sin_family = PF_INET;
			addr.sin_port = port;

			connect(sock, (sockaddr *)&addr, sizeof(sockaddr_in));
		}



};



// should have new common ancestor with socketbinstr

class udpsocketbinstr : public binstream {
	protected:
		int sock;

/*		static dword checkerror(int ret, const char *where = "") {
			if (ret == -1)
				throw wstd_error(where);
			else
				return ret;
		}*/
	public:
		udpsocketbinstr(bool create = true) {
	        if (create)
	        	checkerror(sock = socket(AF_INET, SOCK_DGRAM, 0), "socket");
	        else
	        	sock = -1;
        }

        ~udpsocketbinstr() {
        	if (sock != -1) close(sock);
        }


        int fd() { return sock; }

        void nonblocking(bool block) { checkerror(fcntl(sock, F_SETFL, O_NONBLOCK)); }
        	
		virtual dword write(const void *co, dword size)
			{ return checkerror(TEMP_FAILURE_RETRY(::write(sock, co, size)), "write"); }
		virtual dword read (void *co, dword size)
			{ return checkerror(TEMP_FAILURE_RETRY(::read(sock, co, size)), "read"); }
		virtual dword flush() { return 0; }


		virtual dword writeaddr(const void *co, dword size, const sockaddr_in &addr) {
			return checkerror(TEMP_FAILURE_RETRY(::sendto(sock, co, size, 0, (sockaddr *)&addr, sizeof(addr))), "sendto");
		}

		virtual dword readaddr(void *co, dword size, sockaddr_in &addr) {
			socklen_t addrsize = sizeof(addr);
			
			dword out = checkerror(TEMP_FAILURE_RETRY(::recvfrom(sock, (char *)co , size, 0, (sockaddr *)(void *)&addr, &addrsize)), "recvfrom");
			if (addrsize != sizeof(addr))
				throw std::runtime_error("Wrong size of addr in recvfrom");

			return out;
		}

};

class udpsrvsocketbstr : public udpsocketbinstr {
	public:
		void create(word port, bool listen = false, int backlog = 10) {
			sockaddr_in addr;
			addr.sin_addr.s_addr = INADDR_ANY;
			addr.sin_family = AF_INET;
			addr.sin_port = htons(port);
			checkerror(bind(sock, (sockaddr *)&addr, sizeof(sockaddr_in)), "bind");
			
			if (listen)
				checkerror(::listen(sock, backlog));
		}
		udpsrvsocketbstr() {}
		udpsrvsocketbstr(word port, bool listen = false, int backlog = 10) { create(port, listen); }
		
		void accept(udpsocketbinstr &co, sockaddr_in &addr) {
			socklen_t len = sizeof(addr);
			checkerror(TEMP_FAILURE_RETRY(::accept(  ((udpsrvsocketbstr &)co).sock, (sockaddr *)&addr, &len) ));				
		}
};


class udpclntsocketbstr : public udpsocketbinstr {
	public:

		void create(word port, const char *haddr) {
			sockaddr_in addr;
			hostent *h;
			h = gethostbyname(haddr);

			if (h->h_addr_list[0])
				addr.sin_addr = *(in_addr *)h->h_addr_list[0];
			else
				throw std::runtime_error("No such host");

		   addr.sin_family = AF_INET;
		   addr.sin_port = htons(port);

			checkerror(connect(sock, (sockaddr *)&addr, sizeof(sockaddr_in)), "connect");
	   }

		udpclntsocketbstr()  {}


		udpclntsocketbstr(word port, const char *haddr)  {
			create(port, haddr);
		}

		udpclntsocketbstr(word port, int haddr) {
			sockaddr_in addr;
			//inet_aton(haddr, &addr.sin_addr);
			addr.sin_addr.s_addr = haddr;
			addr.sin_family = AF_INET;
			addr.sin_port = htons(port);

			checkerror(connect(sock, (sockaddr *)&addr, sizeof(sockaddr_in)), "connect");
		}



};




















struct clntinfo {
	int fd;
    in_addr a;

    clntinfo() : fd(-1) {}
    clntinfo(int fd, in_addr co) : fd(fd), a(co) {}
    clntinfo(int fd) : fd(fd) {}

    bool operator==(const clntinfo &co) { return co.fd == fd; }
};

class clntempty {
	public:
    	bool operator() (const clntinfo &co) { return co.fd == -1; }
        void setnotempty(clntinfo *co) { co->fd = 1; }
        void setempty(clntinfo *co) { co->fd = -1; }
        clntinfo *newo(int count) { return new clntinfo[count]; }
};

extern "C" int abs(int);

/*class clnthash {
	private:
    	elc limit;
	public:
    	clnthash(elc limit) : limit(limit) {}
    	elc operator() (const clntinfo &co) { return abs(co.fd) % limit; }
};


inline void FDSET(int fd, fd_set &co) {
	FD_SET(fd, &co);
}
    

/*
class ssrvsocketbstr : public srvsocketbstr, public synchronizer {
	private:
    	int actsock;
        int locksock;
        typedef smphashtbl<lnxvector<clntinfo>, clntempty, clnthash> hashtbl;
        hashtbl clients;

	public:
    	ssrvsocketbstr(word port, elc count) : srvsocketbstr(port), clients(count, clnthash(count)), actsock(-1), locksock(-1) {
        	cout << " Server started: " << '\n';
        }
    
    	virtual void rlock() {
        	locksock = actsock;
        }

    	virtual void runlock() {
        	locksock = -1;
        }

    	virtual void rwlock() {
            rlock();
        }

    	virtual void rwunlock() {
            runlock();
        }

        virtual bool locked() { return actsock == locksock; }

        pollfd makepollfd(int fd, short events) {
        	pollfd f;
            f.fd = fd;
            f.events = events;
            return f;
        }

        virtual dword read(void *co, dword ile) {

        	lnxvector<pollfd> fds(clients.size() + 2);
            bool moreclients = true;

            while (true) {
            	fds[0] = makepollfd(0, POLLIN);
                int count = 1;
                if (moreclients)
                	fds[count++] = makepollfd(sock, POLLIN);
//                if (locksock != -1)
//	                fds[count++] = makepollfd(locksock, POLLIN);
                    
                foreach(hashtbl::itr, i, clients)
                	if (locksock == -1)
						fds[count++] = makepollfd(i->fd, POLLIN | POLLHUP | POLLERR);
                    else
						fds[count++] = makepollfd(i->fd, locksock == i->fd ? (POLLIN | POLLHUP  | POLLERR) : (POLLHUP  | POLLERR));

                                
            	poll(fds, count, -1);

                if (fds[0].revents & POLLIN) {
                	char cmd;
                	::read(0, &cmd, 1);
                    switch (cmd) {
                    	case 'q' : exit(1);
                        case 's' : {
                        	cout << " Clients: \n";
			                foreach(hashtbl::itr, i, clients)
                            	cout << inet_ntoa(i->a) << '\n';
                            break;
                        }
                    }
                }

                                        
    	        if (fds[1].revents & POLLIN) {
                    // new client
                    if (clients.full())
                    	moreclients = false;
                    else {
	                    sockaddr_in addr;
	                    socklen_t size = sizeof(sockaddr_in);
	                    clntinfo c;
		                c.fd = accept(sock, (sockaddr*)&addr, &size);
		                c.a = addr.sin_addr;
	                    clients.ins(c);
		                cout << "Client attached: " << inet_ntoa(c.a) << '\n';
                    }
                }
                	// call from client or detach
                hashtbl::itr j;
                for(lnxvector<pollfd>::itr i = fds.begin() + 2; i != fds.begin() + count; ++i)
                    if (i->revents & (POLLHUP | POLLERR) ) {
                       	close(i->fd);
                        clntinfo a(i->fd);
                        j = clients.find(a);
                        if (j) {
	 	                    cout << "Client detached: " << inet_ntoa(j->a) << '\n';
	                        clients.del(j);
                            moreclients = true;
                        }
                    } else if (i->revents & POLLIN)
                      	return ::read(actsock = i->fd, co, ile);
            }
        }

        virtual dword write(void *co, dword ile) {
        	return ::write(actsock, co, ile);
        }

};

*/
class srvnpipebinstr : public npipebinstr {
	public:
	    srvnpipebinstr() {
		   	mkfifo("/tmp/marservwriterfifo.pipe", 0666);
		   	mkfifo("/tmp/marservreaderfifo.pipe", 0666);

        
		   	reader = open("/tmp/marservwriterfifo.pipe", O_RDWR);
		   	writer = open("/tmp/marservreaderfifo.pipe", O_RDWR);
	    }
};

/*

class clntnpipebinstr : public npipebinstr {
	public:
		clntnpipebinstr() {
			writer = open("/tmp/marservwriterfifo.pipe", O_WRONLY);
			reader = open("/tmp/marservreaderfifo.pipe", O_RDONLY);
		}
};



class msgbinstr : public binstream {
	private:
    	int msgidread;
        int msgidwrite;

    protected:
        int wmsgtype;
        byte wbuf[200];
        byte *wpos;

    protected:
        int rmsgtype;
        byte rbuf[200];
        byte *rpos;
        word rsize;
        
	public:
    	msgbinstr(bool server = false) : wpos(wbuf), rpos(rbuf), rsize(0) {
        	if (server) {
	        	msgidread  = msgget(0xAF0BE176, 0666);
	        	msgidwrite = msgget(0xAF0BE177, 0666);

                msgctl(msgidread, IPC_RMID, 0);
                msgctl(msgidwrite, IPC_RMID, 0);

	        	msgidread = msgget(0xAF0BE176, 0666 | IPC_CREAT);
	        	msgidwrite = msgget(0xAF0BE177, 0666| IPC_CREAT);
            } else {
	        	msgidread = msgget(0xAF0BE177, 0666| IPC_CREAT);
	        	msgidwrite = msgget(0xAF0BE176, 0666| IPC_CREAT);
            }
        }
        ~msgbinstr() {
	      	//msgctl(msgidread , IPC_RMID, 0);
	       	//msgctl(msgidwrite, IPC_RMID, 0);
        }
            
    
		virtual dword write(void *co, dword size) {
			if (co) memcpy(wpos, co, size);
			wpos +=size;
		}
		virtual dword read(void *co, dword size) {
			if (rpos - rbuf == rsize) {
				int res;
				do {
					res =  msgrcv(msgidread, &rmsgtype, 200, rmsgtype, 0) ;
				} while (res == -1);

				rsize = res - 4;
				rpos = rbuf;
			}
			if (co) memcpy(co, rpos, size);
			rpos += size;
		}
		virtual dword flush() {
			msgsnd(msgidwrite, &wmsgtype, wpos - wbuf + 4, 0);
			wpos = wbuf;
		}

};

  */

/*
//template<class C>
class clntmsgbstrs : public msgbinstr, public synchronizer {
	private:
    	bool flocked;
        typedef msgbinstr inherited;
	public:
    
        clntmsgbstrs() { wpos = wbuf + 2; }
        
		virtual void rlock() {
        	if (flocked) return;
        	wmsgtype = 1;
            wbuf[0] = 'r';
            wbuf[1] = 'l';
            *(int *)(wbuf + 2) = getpid();
            msgbinstr::write(0, 4);
            msgbinstr::flush();
            wpos = wbuf + 2;
            
            rmsgtype = wmsgtype = getpid();
            msgbinstr::read(0, 2);
            flocked = true;
        }
		virtual void runlock() {
        	if (!flocked) return;
            wbuf[0] = 'r';
            wbuf[1] = 'u';
            *(int *)(wbuf + 2) = getpid();
            msgbinstr::write(0, 4);
            msgbinstr::flush();
            wpos = wbuf + 2;
            
            rmsgtype = getpid();
            msgbinstr::read(0, 2);
            flocked = false;
        }

		virtual bool locked() { return flocked; }

		virtual void rwlock() { rlock(); }
		virtual void rwunlock() { runlock(); }

        // ***************************** /

		virtual dword flush() {
        	wbuf[0] = 'd';
        	wbuf[1] = 't';
            msgbinstr::flush();
            wpos = wbuf + 2;
        }
};



class srvmsgbinstr : public msgbinstr {
	private:
    	int pid;
	public:
        srvmsgbinstr() : msgbinstr(true), pid(1) { }


        virtual dword read(void *co, dword size) {
        	do {
           	while (pid == 1) {
               	rmsgtype = 1;
                msgbinstr::read(0, 6);
   	            if (rbuf[0] == 'r' && rbuf[1] == 'l') {// read lock
                	rmsgtype = wmsgtype = pid = *(int*)(rbuf+2);
                    *(word*)wbuf = *(word *)rbuf;
                    msgbinstr::write(0, 2);
                    flush();
                }
            }
           	if (rpos - rbuf == rsize) { // new message will be received
           		msgbinstr::read(0, size + 2);
                if (rbuf[0] == 'd' && rbuf[1] == 't') {
                  	if (co) memcpy(co, rbuf + 2, size);
                } else if (rbuf[0] == 'r' && rbuf[1] == 'u') {
                    pid = rmsgtype = 1;
                    *(word *)wbuf = *(word *)rbuf;
                    msgbinstr::write(0, 2);
                    flush();
                }
            } else
              	msgbinstr::read(co, size);
            } while (pid == 1);
		}
};

*/

/*
class semsynchro : public synchronizer {
	private:
    	int semid;
		bool flocked;
	public:
		virtual void rlock() {
        	if (flocked) return;
        	sembuf b;
            b.sem_num = 0;
            b.sem_op = -1;
            b.sem_flg = 0;
            
            semop(semid, &b, 1);
            flocked = true;

        }
		virtual void runlock() {
        	if (!flocked) return;
        	sembuf b;
            b.sem_num = 0;
            b.sem_op =  1;
            b.sem_flg = 0;
            
            semop(semid, &b, 1);
            flocked = false;
        }
		virtual bool locked() { return flocked;}

		virtual void rwlock() { rlock(); }
		virtual void rwunlock() { runlock(); }
        

		~semsynchro() {}
		semsynchro(bool server = false):  flocked(false) {
            if (server) {
	 	      	semid = semget(0xAF0BE175, 1, 0666);
                semctl(semid, 0, IPC_RMID, 0);
            }
        	semid = semget(0xAF0BE175, 1, 0666 | IPC_CREAT | IPC_EXCL);

            if (semid != -1 || server) { // new sem - set value
            	typedef union {
                	int val;
                } semun;
            	semun arg;
                arg.val = 1;
                semctl(semid, 0, SETVAL, arg);
            } else
	        	semid = semget(0xAF0BE175, 1, 0666);
        }
};
  */
/*
class shrdmem {
	protected:
    	void *m;

    // ...
};

class lnxshrdmem : public shrdmem {
	private:
		int shmid;
	public:
    	lnxshrdmem(key_t key, dword size = 0, int perm = 0666) {
        	shmid = shmget(key, size, 0666 | IPC_CREAT);
            if (shmid == -1)
            	cout << strerror(errno) << endl;
            m = shmat(shmid, 0, 0);
        }
        ~lnxshrdmem() { shmdt(m); }

        void remove() { shmdt(m); shmctl(shmid, IPC_RMID, 0); }

        operator void*() const { return m; }
        
};


        

  */




} //namespace stm

#endif

