#define _FILE_OFFSET_BITS 64
#define _POSIX_REALTIME_SIGNALS

#include <queue>

#include "serverimp.hpp"
#include "stm.hpp"
#include "binstrs.hpp"
#include <dirent.h>
//#include <linux/dirent.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <aio.h>
#include <signal.h>


namespace dftpserver {
using namespace fserver;


using stm::checkerror;

template<class T>
inline T field(const char *buf) { return *(T *)(buf); }

template<class T>

inline void setfield(const char *buf, const T& val) { *(T *)(buf) = val; }


template<>
inline dword field<dword>(const char *buf) {
	return ntohl(*(dword *)(buf));
}

template<>
inline qword field<qword>(const char *buf) {	
	return ( (qword)ntohl(*(dword *)(buf + 4)) )<<32 | ntohl(*(dword *)(buf));
}



template<>
inline void setfield<dword>(const char *buf, const dword &co) {
	 *(dword *)(buf) = htonl(co);
}

template<>
inline void setfield<qword>(const char *buf, const qword &co) {	
	*(dword *)(buf + 4) =  htonl( co >> 32 );
	*(dword *)(buf) =  htonl( co );
}





#define strfld(buf, name) atndptr<char> name( field<dword>(buf) );  \
	memcpy(name, buf + 4, field<dword>(buf));      \
	name[field<dword>(buf)] = 0

	
class netresponse : public responsebuf {
	public:
	    void output(dword co) {
	    	responsebuf::output(htonl(co));
	    }
	    void output(qword co) {
	    	responsebuf::output( ((qword)htonl(co) << 32) |  (htonl(co << 32)) );
	    }
	
	    void output(int co) {
	    	responsebuf::output(htonl(co));
	    }
	    void output(long co) {
	    	responsebuf::output(htonl(co));
	    }
	    void output(size_t co) {
	    	responsebuf::output(htonl(co));
	    }
	
	
		
		void output(const string &co) {
			output(co.length());
			add(co.c_str(), co.length());
		}
		void output(const char *co) { add(co, strlen(co)); }
		void output(char *co) { add(co, strlen(co)); }



//		template<class T>
//			void output(T co) { add((const char *)(void *)&co, sizeof(T) ); }

		netresponse(const netresponse &co) : responsebuf(co) {}
		netresponse()  {}
		
		


};
	

class errorresponse : public netresponse {
	public:
		errorresponse(dword id, const char *msg) {
			output(10);
			output(id);
			output(msg);
		}
};



class dftpidpacket : public idpacket {
	protected:
		virtual dword id(const char *co, dword count) const { return field<dword>(co + 4); }
		
		void checklogged(connection *conn) {
			if (!conn || !conn->authorized())
				throw server_error("Authorization required");
		}
		//void handle(connection *conn, const char *co, dword count) {
		//	idpacket::handle(conn, co, count);
		//}

		
	public:
		dftpidpacket(server *s, dword type) : idpacket(s, type) {}
		
	
};

class hellopacket : public packet {
	public:
		hellopacket(server *s) : packet(s, 0) {}

		void handle(connection *conn, const char *co, dword count) {
			responsebuf *r = new responsebuf();
			r->output(0);
			r->output(1);
			//r.output(serv->welcome.length());
			r->output(serv->welcome);

			serv->answer(r);
		}
};


class authpacket : public dftpidpacket {
	public:
		authpacket(server *s) : dftpidpacket(s, 1) {}

		void handle(connection *conn, const char *co, dword count) {
			dftpidpacket::handle(conn, co, count);
			
			if (!conn)
				conn = addconn();

			strfld(co + 8, name);
			strfld(co + 12 + field<dword>(co + 8), pswd);


			if (!conn->authorize(name, pswd))
				conn->answer(new errorresponse(field<dword>(co + 4), "No such user") );

		}
};


class logoutpacket : public dftpidpacket {
	public:
		logoutpacket(server *s) : dftpidpacket(s, 2) {}

		void handle(connection *conn, const char *co, dword count) {
			dftpidpacket::handle(conn, co, count);
			serv->st()->removeconn(conn);
		}

};


class dirlstresponse : public response {
	protected:
		dword id;
				
		static const word packsize = 50000;
		list<netresponse> resps;
		
		void addrec(const char *name, const struct stat& s) {
			netresponse *resp = &resps.back();
			
			if (resps.empty() ||
				field<dword>(*resp + 8) + 4 + strlen(name) + 6*4 > packsize )
			{
				resps.push_back(netresponse());
				resp = &resps.back();
				
				resp->output(11u);
				resp->output(id);
				resp->output(0);
			}
			
			setfield<dword>(*resp + 8, field<dword>(*resp + 8) + 4 + strlen(name) + 6*4);
			
			
    		resp->output(strlen(name));
    		resp->output(name);

    		resp->output(s.st_uid);
    		resp->output(s.st_gid);
    		resp->output(s.st_mode);
    		resp->output(s.st_atime);
    		resp->output(s.st_mtime);
    		resp->output(s.st_ctime);
  		}
		
	public:
		dirlstresponse(dword id, const char *dirname);
		
		virtual void send(stm::binstream *s) {
			foreach(list<netresponse>::iterator, i, resps)
				i->send(s);				
		}
		
};

dirlstresponse::dirlstresponse(dword id, const char *dirname) : id(id) {
	DIR *d(opendir(dirname));
			
	if (!d)
		throw stm::wstd_error("Dir list");
				
	dirent *dir;
	struct stat s;
	dword count = 0;
	string strdir = dirname;
	
	if (strdir[strdir.length()-1] != '/')
		strdir += "/";
			

	while (dir = readdir(d)) {
		int out;
		string realdir = strdir + dir->d_name;
			
		if ((out = stat(realdir.c_str(), &s)) == -1)
			continue;

		addrec(dir->d_name, s);

	}
}


class listpacket : public dftpidpacket {
	public:
		listpacket(server *s) : dftpidpacket(s, 3) {}

		void handle(connection *conn, const char *co, dword count) {
			dftpidpacket::handle(conn, co, count);
			
			checklogged(conn);
			
			strfld(co + 8, dir);
			
			int uid = geteuid();
			int gid = getegid();
   			int out;
			try {							
				if (seteuid(conn->uid) == -1)
					throw stm::std_error();
				
				if (seteuid(conn->gid) == -1)
					throw stm::std_error();
					
								    				
    			if ( (out = open(dir, O_RDONLY)) == -1 )
    				if (errno != EACCES)
	    				throw stm::code_error(out);	    		
	    			else
	    				close(out);	    				    					
    		} finally ( { stm::checkerror(seteuid(uid)); stm::checkerror(setegid(gid)); } )
    		
    		if (out == -1)
				conn->answer(new errorresponse (field<dword>(co + 4), "Permission denied"));
			else
				conn->answer(new dirlstresponse (field<dword>( co + 4), dir));
		}

};



// findcon and reqest must throw when not found

/*		
class afilerequest : public request {
	protected:
		aiocb cb;
		
		
		void read() {
			checkerror(aio_read(&cb));
		}
		
		void write() {
			checkerror(aio_write(&cb));
		}

	public:
		virtual bool notifyended();

		afilerequest(connection *conn, dword id, const char *file) : request(conn, id) {
			cb.aio_fildes = checkerror(TEMP_FAILURE_RETRY(open(file, O_RDONLY)));
			
			cb.aio_sigevent.sigev_notify = SIGEV_SIGNAL;
			cb.aio_sigevent.sigev_signo = SIGUSR1;
			cb.aio_sigevent.sigev_value.sival_ptr = this;

		}

		virtual ~afilerequest();
	public:
		virtual void dummy();
					
};


afilerequest::~afilerequest() {  // g++ jest zalosny
	checkerror(aio_cancel(cb.aio_fildes, &cb));
	checkerror(close(cb.aio_fildes));
}

void afilerequest::dummy() { id + id;}  // tu tez

  */
/*
class asyncbstrhandle;

class asyncbinstr : public binstream {

	public:
		virtual asyncbstrhandle *aread(offs_t offset, void *buf, dword count) = 0;
		virtual asyncbstrhandle *awrite(offs_t offset, const void *buf, dword count) =0;
		
		virtual asyncbstrhandle *aread(offs_t offset, void *buf, dword count, void ) = 0;
		virtual asyncbstrhandle *awrite(offs_t offset, const void *buf, dword count) =0;
		

};


class asyncfilebinstr : public binstream {
	protected:
		aiocb cb;

	public:
		virtual dword read(void *buf, dword count) { checkerror(TEMP_FAILURE_RETRY(read(cb.aio_filedes, buf, count))); }
		virtual dword write(const void *buf, dword count) { checkerror(TEMP_FAILURE_RETRY(write(cb.aio_filedes, buf, count))); }


};

  */

// you have to guarantee the buffer is not destroyed until the data is send
class persistentbufresponse : public response {
	protected:
		const char *buf;
		dword count;
	public:
		persistentbufresponse(const char *buf, dword count)
			: buf(buf), count(count)
		{}
				

	
		virtual void send(stm::binstream *co) {
			co->write(buf, count);		
		}
};

class sendfilerequest : public request {
	protected:
		aiocb cb;
		
		
		void read() {
			checkerror(TEMP_FAILURE_RETRY(aio_read(&cb)));
		}
		
		void write() {
			//checkerror(TEMP_FAILURE_RETRY(aio_write(&cb)));
			conn->answer(new persistentbufresponse((const char *)cb.aio_buf, cb.aio_nbytes), this);
		}
	
	protected:
		char buf [1 << 16];

		qword offset;


		qword count;

		//int filedes;


		dword readcount;

		static const dword packsize = (1 << 16) - 20 - 10000;
		
		
		
		
	

		void setnextread() {
			//cb.aio_fildes = filedes;
			if (count) {
				cb.aio_buf = buf + 8 + 8 + 4;
				cb.aio_offset = offset;
				readcount = cb.aio_nbytes = count > packsize ? packsize : count;
				count -= cb.aio_nbytes;
				offset += cb.aio_nbytes;

				state = reading;
			} else {
				state = done;
				remove = true;
				
			}
		}


		void setwrite() {
			setfield<dword>(buf, 12);
			setfield<dword>(buf + 4, id);
			setfield<qword>(buf + 8, cb.aio_offset);

			cb.aio_buf = buf;

//			cb.aio_fildes = conn->sock.fd();
			cb.aio_nbytes = readcount + 20;
			state = writing;
		}




	public:
		enum { reading, writing, done, destructing, cancelling  } state;

		virtual bool notifysignal() {
			if (state == destructing) return false;
			if (state == cancelling) {
				state = done;
				remove = true;	// or delete this;		
				return false;
			}
			try {

				int err = TEMP_FAILURE_RETRY(aio_error(&cb));
				if (err == EINPROGRESS || err == ENOSYS || err == EINVAL)
					throw stm::wcode_error(err, "Asynchronous io operation");

				if (err != 0)
					throw stm::wcode_error(err, "Asynchronous io operation");

				conn->serv()->subscribewrite(this); // wait for writing		
				
				return true;
			} catch (exception &e) {
				// reentrant?
				conn->answer(new errorresponse(id, e.what()));
				cerr << e.what();
			} catch (...) {
				cerr << "Unknown exception caught in signal handler \n";
			}
			return false;
		
		}
		virtual bool notifywrite() {
			if (state == destructing || state == done) return true;
			if (state == cancelling) {
				state = done;
				remove = true;	// or delete this;		
				return true;
			}
			setwrite();
			write();
			return true;
		}
		
		virtual void notifywritten(response *co) {
			setnextread();
			if (state == reading)
				read();												
		}




		sendfilerequest(connection *conn, dword id, const char *file, qword offset, qword count)
			: request(conn, id), offset(offset), count(count)

		{
			cb.aio_fildes = checkerror(TEMP_FAILURE_RETRY(open(file, O_RDONLY)));
			cb.aio_reqprio = 0;
			
			cb.aio_sigevent.sigev_notify = SIGEV_SIGNAL;
			cb.aio_sigevent.sigev_signo = SIGUSR1;
			cb.aio_sigevent.sigev_value.sival_ptr = this;
			
			
			assert(file);
			assert(conn);

			//filedes = cb.aio_fildes;

			setnextread();
			if (state == reading) read();
		}
		~sendfilerequest() {

			if (state == writing) {  // set filedes back
				state = destructing;
				checkerror(aio_cancel(cb.aio_fildes, &cb));
//				cb.aio_fildes = filedes;
			} else
				state = destructing;
		}



};


class getfilepacket  : public dftpidpacket {
	public:
		getfilepacket(server *s) : dftpidpacket(s, 4) {}

		void handle(connection *conn, const char *co, dword count) {
			dftpidpacket::handle(conn, co, count);
			
			checklogged(conn);
			
			strfld(co + 8, file);

			
			struct stat s;
			
			
			
			int uid = geteuid();
			int gid = getegid();
   			int out;
			try {			
				if (seteuid(conn->uid) == -1)
					throw stm::std_error();
				
				if (seteuid(conn->gid) == -1)
					throw stm::std_error();
				
			
    			if ( (out = stat(file, &s)) == -1 )
    				if (errno != EACCES)
	    				throw stm::code_error(out);
	    				
    			if ( (out = open(file, O_RDONLY)) == -1 )
    				if (errno != EACCES)
	    				throw stm::code_error(out);	    		
	    			else
	    				close(out);
	    				
    				
    		} finally ( { stm::checkerror(seteuid(uid)); stm::checkerror(setegid(gid)); } )
    						

	/*		if (conn->uid == s.st_uid && s.st_mode&S_IRUSR ||
				conn->gid == s.st_gid && s.st_mode&S_IRGRP ||
				s.st_mode&S_IROTH*/
			if (out != -1) {
				qword count = field<qword>(co + 12 + field<dword>(co + 8) + 8);
				if (!count)
					count = s.st_size;
					
				sendfilerequest *r = new sendfilerequest(
					conn,
					id(co, count),
					file,
					field<qword>(co + 12 + field<dword>(co + 8)),
					count

				);
				
			} else
				conn->answer(new errorresponse (field<dword>(co + 4), "Insufficient rights to get the file"));

				

		}


};



class cancelfilepacket : public dftpidpacket {
	public:
		cancelfilepacket(server *s) : dftpidpacket(s, 5) {}


		void handle(connection *conn, const char *co, dword count) {
			dftpidpacket::handle(conn, co, count);
			checklogged(conn);
			request *r = getrequest(conn, co, count);
//			if (dynamic_cast<sendfilerequest *>(r))  // it returns 0, but sendfilerequest is the only non-abstract descendant of request
			if (static_cast<sendfilerequest *>(r))
				if ( ((sendfilerequest *)r)->state != sendfilerequest::done )
					((sendfilerequest *)r)->state = sendfilerequest::cancelling;
				else
					delete r;																	
			 else
				throw server_error("Wrong request number");
						
		}
};
		


class reqsighandler {
	protected:
		static reqsighandler *current;
		static void handler(int sig, siginfo_t *si, void *) {
			if (current) current->fire(si->si_value.sival_ptr);
		}

		int signo;
		
		

	public:

		reqsighandler(int sig) : signo(sig)  {
			current = this;

			struct sigaction act;
			act.sa_sigaction = handler;
			sigfillset(&act.sa_mask);
			act.sa_flags = SA_SIGINFO;

			checkerror(sigaction(sig, &act, 0));

		}

		~reqsighandler() {
			current = 0;
			struct sigaction act;
			act.sa_handler = SIG_DFL;
			sigfillset(&act.sa_mask);
			act.sa_flags = 0;

			checkerror(sigaction(signo, &act, 0));

		}
	protected:

		virtual void fire(void *r) {}
			// theoretically, only afilerequests can raise the signal
//			sendfilerequest *req = dynamic_cast<sendfilerequest *>(static_cast<request *>(r));
			
			

//			req->notifyended();

/*			foreach(state::connset::iterator, i, currentstate->connections)
				foreach(connection::requestset::iterator, j, (*i)->requests) {
					afilerequest *r = dynamic_cast<afilerequest *>(*j);

					if (r)
						if (r->notifyended())
							return;
				}                  */
};


class qreqsighandler : public reqsighandler {
	public:
		queue<request *> reqq;
		
		
		qreqsighandler(int sig) : reqsighandler(sig) {}
		
		
		void lock() {	
			sigset_t mask;
			sigemptyset(&mask);
			sigaddset(&mask, signo);
			
			stm::checkerror(sigprocmask(SIG_BLOCK, &mask, 0));
				
		}
		
		void unlock() {	
			sigset_t mask;
			sigemptyset(&mask);
			sigaddset(&mask, signo);
			
			stm::checkerror(sigprocmask(SIG_UNBLOCK, &mask, 0));
				
		}
	protected:
		
		virtual void fire(void *r) {
			reqq.push(static_cast<request *>(r));
		}
		
	
};





static adptr<qreqsighandler> sighandler;
reqsighandler *reqsighandler::current = 0;


void notifysignal() {
	
	while (!sighandler->reqq.empty()) {
		sendfilerequest *r = 0;
		try {
			sighandler->lock();
			r = dynamic_cast<sendfilerequest *>(sighandler->reqq.front());
			sighandler->reqq.pop();
		} finally ( { sighandler->unlock(); } )
		
		if (r) r->notifysignal();
	}
}


void senderror(server *s, int id, const char *msg) {
	try {
		s->answererror(new errorresponse(id, msg));
	} catch (stm::std_error &e) {
		cerr << "Message \"" << msg << "\" could not been delivered to client \n";
		cerr << "because of the following reason: " << e.where() << " : " << e.what() << "\n";
	} catch (std::exception &e) {
		cerr << "Message \"" << msg << "\" could not been delivered to client \n";
		cerr << "because of the following reason: " << e.what() << "\n";
	} catch (...) {
		cerr << "Message \"" << msg << "\" could not been delivered to client \n";
	}
}

dword packettype(const char *buf, dword count) {
	if (count < 4)
		return 0xFFFFFFFF;
	else
		return field<dword>(buf);
}

void registerdftpprotocol(server *s) {
	//delete new afilerequest(0, 1, 0);
	s->regpacket(new hellopacket(s));
	s->regpacket(new authpacket(s));
	s->regpacket(new logoutpacket(s));
	s->regpacket(new listpacket(s));
	s->regpacket(new getfilepacket(s));
	s->regpacket(new cancelfilepacket(s));

	sighandler = new qreqsighandler(SIGUSR1);

	s->senderrorfunc = senderror;
	s->notifysignalfunc = notifysignal;
	s->beforepoll = notifysignal;
	s->packettype = packettype;
}


} // namespace fserver
