#ifndef _MAR_PARAMPARSER_HPP
#define _MAR_PARAMPARSER_HPP

//#define _GNU_SOURCE
#include <getopt.h>
#include <unistd.h>
#include <memory.h>

#include "s.h"

#include <exception>
#include <stdexcept>


namespace mar {
//
class paramsparser {
	private:
		char * const * argv;
		int argc;
		const char *optstr;
		option *longopts;


	friend class itr;
	public:
		class unknownoption : public std::runtime_error {
			public:
			virtual const char* what () const throw() {
				return "Unknown option";
			}
		};

		class missingargument : public std::runtime_error {
			public:
			virtual const char* what () const throw() {
				return "Missing argument" ;
			}
		};
	protected:
		char getopt(int &longoptindex) {
			char out = ::getopt_long(argc, argv, optstr, longopts, &longoptindex);
			switch (out) {
				case ':' :
					throw missingargument();

				case '?' :
					throw unknownoption();
				case -1 :
					longoptindex = -1;
					return 0;
			}
			return out;
		}
	public:



		paramsparser(int argc, char * const *argv, const char *optstr, option *longopts)
			: argv(argv), argc(argc), optstr(optstr), longopts(longopts)
		{
			opterr = 0; // redudancy, I know, but if I did static object, I would have to add .cpp file
						// and it would only make things harder
		}


		class itr {
			protected:
				paramsparser *pp;
				int longopt;
				char opt;

			public:
				itr(paramsparser *co) : pp(co), opt(0), longopt(-1) {}

				itr &operator++() {
					opt = pp->getopt(longopt);
					return me;
				}
				bool operator==(const itr &co) { return !memcmp(this, &co, sizeof(itr)); }
				bool operator!=(const itr &co) { return !(co.pp == pp && co.longopt == longopt && co.opt == opt); /*memcmp(this, &co, sizeof(itr));*/ }

				const char *operator *() const { return opt ? &opt : (longopt != -1 ? pp->longopts[longopt].name : 0); }
				const char *operator ()() const { return operator*(); }

				const char *arg() const { return optarg; }
				bool longoption() const { return opt; }

				
		};

		itr begin() { return ++(itr(this)); }
		itr end() { return itr(this); }

		char *const *notoptbegin() { return argv + optind; }
		char *const *notoptend() { argv + argc; }

		// this is truely speaking very ugly implementation - you cannon use ntoptbegin before traversing usual iterator
		// you also cannot have more than one iterator and so on ...

		// to do it well, you would have to not use getopt, but I have no time to write it from scratch

};


} //namespace mar

using namespace mar;

#endif

