//---------------------------------------------------------------------------
#pragma hdrstop

#include "serverimp.hpp"

namespace fserver {



request::~request() { conn->removereq(this); }
void request::dummy() { }



//**************************************  C O N N E C T I O N *******************************

request *connection::getrequest(word id) {
	access();
	request *out = 0;
	foreach(requestset::iterator, i, requests)
		if ( (*i)->id == id)
			out = *i;
		else if ( (*i)->remove == true) {
			requestset::iterator t = i;
			--i;
			requests.erase(t);
		}
		
	if (!out)
		throw server_error("No such request");
		
	return out;
}

connection::~connection() {
	requestset temp;
	temp.swap(requests);  // to prevent requests removing themselves from the list
	foreach(requestset::iterator, i, temp)
 		delete *i;
}



//**************************************  S T A T E *******************************


state::connset::iterator state::findconn(const sockaddr_in &addr) {
	connset::iterator out = 0;
	time_t time = ::time(0);

	foreach(connset::iterator, i, connections)
		if (!memcmp( &(*i)->client, &addr, sizeof(addr) ))
			out = i;
		else if (time - (*i)->lastaccess > sessiontimeout) {
			connset::iterator t = i;
			--i;
			connections.erase(t);
		}

	if (out != 0 && *out) (*out)->access();
	return out;
}



void state::removeconn(const sockaddr_in &addr) {
	connset::iterator c = findconn(addr);
	connection *t;

	if (c != 0) {
		t = *c;
				
		serv->connremoved(t, t->idx());
		connections.erase(c);
		delete t;
	}
}
void state::removeconn(connection *c) {
	assert(c);
	if (!connections.empty()) {
		connections.erase(connections.begin() + c->idx());
	}
				
	serv->connremoved(c, c->idx());
	delete c;
}
connection *state::addconn(const sockaddr_in &addr) {
	connection *out = new connection(this, addr);
	connections.push_back(out);
	serv->connadded(out, connections.size() - 1);
	return out;
}

connection *state::addconn(connection *co) {
	connections.push_back(co);
	serv->connadded(co,  connections.size() - 1);
	return co;
}
		
connection *state::addconn() {
	connection *out = new connection(this);
	connections.push_back(out);
			
	serv->connadded(out,  connections.size() - 1);
			
	return out;
}



//**************************************  S E R V E R *******************************


server::~server() {
	//sockets.erase(sockets.begin(), sockets.end());		// to prevent deleting it by connections each after another
	
	
	packetset temp;
	temp.swap(packets);
	
	cout << packets.empty() << (packet *)*packets.begin() << endl;
	foreach(packetset::iterator, i, temp) {
		cout << (*i)->serv->welcome << endl;
		++i;		
		cout << (*i)->serv->welcome << endl;
		--i;
		
		delete (packet *)(*i);
		
		++i;		
		cout << (*i)->serv->welcome << endl;
		--i;
		
	}
}


packet *server::findpacket(dword type) const {
	packetset::const_iterator i;
	for(i = packets.begin(); i != packets.end(); ++i)
		if ((*i)->type == type)
			break;
	if (i == packets.end())
		throw server_error("Unrecognized packet type");
	else
		return *i;
}


void server::sendresponseto(response *r, const sockaddr_in &towho) {
   	try {
   		memcpy(&who, &towho, sizeof(who));
   		stm::checkerror( connect(sock.fd(), (sockaddr *)&who, sizeof(who)) );
   		r->send(&sock);			
			
   	} catch (...) {
   		sockaddr_in adr;
   		adr.sin_family = AF_UNSPEC;
   		stm::checkerror( connect(sock.fd(), (sockaddr *)&adr, sizeof(adr)) );
   		who.sin_addr.s_addr = 0;
   		
   		throw;
   	}

   	sockaddr_in adr;         // unconnect
   	adr.sin_family = AF_UNSPEC;
   	stm::checkerror( connect(sock.fd(), (sockaddr *)&adr, sizeof(adr)) );
   	who.sin_addr.s_addr = 0;
   	

}



bool server::sendresbuf() {
	if (responsebuf.empty())
		return false;
	else {
		try {
			sendresponseto(*responsebuf.begin(), responsebuf.begin()->who);
						
		} catch (stm::std_error &e) {
			if (e.code != EAGAIN)
				throw;		
			else
				return true;
		}
		
		
		response *r = *responsebuf.begin();
	 	request *notify = responsebuf.begin()->notify;

		responsebuf.pop_front();
				
		try {
			if (notify) notify->notifywritten(r);
		} finally ( { delete r; } )
		
		return true;
	}
	
					
//	foreach(responseset::iterator, i, responsebuf)
//		(*i)->send(sock);
}

void server::passwrite() {
//	try {   // not needed , sent in main loop
//    	foreach(requestset::iterator, i, notifywrite)
		if ( !notifywrite.empty() ) {
			try {
    			currentreqid = (*notifywrite.begin())->id;
    			currentconn = (*notifywrite.begin())->conn;
    			
        		if ( (*notifywrite.begin())->notifywrite() )
        		 	notifywrite.pop_front()	;
    			
    				
   			} finally ( { currentreqid = 0; currentconn = 0; } )
				
			
				
		}

//	} catch (std_error &e) {  // just in case, answer has it too, but what with rollback? - not needed, whole packet or none
//		if (e.code != EAGAIN)
//			throw;		
//	}
							
	
}


void server::passsignal() {
	if (notifysignalfunc) notifysignalfunc();
}


void server::serve() {
	if (!packettype)
		throw server_error("Cannot serve with no packettype function");

	pollfd pfd = { sock.fd(), POLLIN | POLLOUT, 0} ;
	
	sock.nonblocking(true);
	

	do {
	
	try {
	
		if (beforepoll) beforepoll();
		if (!responsebuf.empty() || !notifywrite.empty())
			pfd.events = POLLIN | POLLOUT;
		else
			pfd.events = POLLIN;
			
		int pollout = poll(&pfd, 1, 1000); // if signal came after call to beforepoll
		
		if (pollout == -1)
			if (errno == EINTR)
				passsignal();
			else
				throw stm::std_error();
				
		else {
			if (pfd.revents & POLLOUT)
    			try {
    				if (!sendresbuf()) 				// first our buffer
	    				passwrite();
    			} catch (stm::std_error &e) {
    				if (e.code != EAGAIN)
    					throw;
    			}
			if (pfd.revents & POLLIN) {
    			dword count = sock.readaddr(buf, bufsize, who);
    			
        		if (count == 0) continue;

        		packet *p = findpacket(packettype(buf, count));
        		
        		currentconn = st()->getconn(who);  // may be 0
        		

        		if (p)
        			p->handle(currentconn, buf, count);
        		else
        			throw server_error("Unrecognized packet type");
       		}
       }
			

	} catch (stm::std_error &e) {
		string msg = string(e.where()) + " : " + e.what();
		cerr << msg << "\n";
		senderror(currentreqid, msg.c_str());   // do exception filter or store id of request   -the latter is better
	} catch (std::exception &e) {
		cerr << e.what();
		senderror(currentreqid, e.what());
	}
	
	currentconn = 0;
	currentreqid = noreq;
	who.sin_addr.s_addr = 0;
	
	} while (true);

}

// cut to 64K?

void server::answer(response *co) {
	if (currentconn != 0)
		sendto(co, currentconn->client);
	else if (who.sin_addr.s_addr != 0)
		sendto(co, who);		
	else {
		const char *err = "No client to send following response:\n";
		stderrbstr.write(err, strlen(err));
		co->send(&stderrbstr);                 // what with protocol?
		delete co;
	}
}


void server::sendto(response *co, const sockaddr_in &who, request *notify) {
	responsebuf.push_back(responseaddr(co, who, notify));
}



void server::answererror(response *co) {
	sendresponseto(co, who);
}



} // namespace fserver

