//---------------------------------------------------------------------------
#ifndef serverimpH
#define serverimpH

#include "s.h"
#include <set>

//#include <vector.hpp>
#include <vector>
#include "binstrs.hpp"
#include <list>
#include <algorithm>

#include <pwd.h>


namespace fserver {

class server_error : public runtime_error {
	public:
		server_error(const char *co) : runtime_error(co) {}
};

class response  {
	public:
		response()  {}
		virtual ~response() {}
		
		virtual void send(stm::binstream *co) = 0;
};


class responsebuf : public response, private std::vector<char> {
	protected:

		virtual void add(const char *co, dword s) {
			insert(end(), co, co + s);
			//memcpy(end() - s, co, s);
		}


	public:
		responsebuf()  {}
		
		virtual ~responsebuf() {}
		
		dword count() { return size(); }
		

		// explicitly defined functions should take precedence over the template generated ones
		// but they don't :(
		void output(const string &co) {
			output(co.length());
			add(co.c_str(), co.length());
		}
		void output(const char *co) { add(co, strlen(co)); }
		void output(char *co) { add(co, strlen(co)); }


//		void output(dword co) { add((const char *)(void *)&co, sizeof(co)); }
//		void output(dword co) { add((const char *)(void *)&co, sizeof(co)); }
//		void output(dword co) { add((const char *)(void *)&co, sizeof(co)); }
//		void output(dword co) { add((const char *)(void *)&co, sizeof(co)); }

	#ifndef __CYGWIN__
		template<class T>
			void output(T co) { add((const char *)(void *)&co, sizeof(T) ); }
	#else
		void output(dword co) { add((const char *)(void *)&co, sizeof(co)); }
	#endif




		operator char *() { return begin(); }
		
		char *operator+(dword co) { return begin() + co; }
		
		
		virtual void send(stm::binstream *co) {
			if (co->write(begin(), size()) != size())
				throw server_error("Not whole response sent");
		}
};


class state;
class server;
class connection;



//class greater<request> {
//	public:
//		bool operator()(const request&co1, const request &co2) { return co1.id > co2.id ? 1 : co1.id == co2.id ? 0 : -1; }
//};

class request {
	public:
		dword id;
		connection *conn;
		
		bool remove;

		bool operator>(const request &co1) const { return id > co1.id; }
		//bool operator==(const request &co1) const { return id > co1.id; }

		request(connection *c, dword id);
		virtual ~request();
		
		virtual bool notifywrite() { return false; }
		virtual void notifywritten(response *co) {  }
		
	protected:
		virtual void dummy();
		

};



class connection {
	friend class state;
	friend class request;
	protected:
		mutable time_t lastaccess;		
		
		void access() const {
			lastaccess = time(0);
		}

	public:
		state *st;
		sockaddr_in client;
		int uid, gid;
		
		int idx();
		
		

		//stm::udpclntsocketbstr sock;

		typedef list<request *> requestset;
		requestset requests;

		request *getrequest(word id);
		connection(state *s) : st(s), uid(-1) {}

		connection(state *s, const sockaddr_in &who)
			: st(s), client(who)/*, sock( ntohs(who.sin_port) , who.sin_addr.s_addr)*/
		{
//			stm::checkerror(connect(sock.fd(), (sockaddr *)&who, sizeof(who)));
//			access();
//			send(sock.fd(), "cos tam cos tam", 16, 0);
//			sendto(sock.fd(), "sendto", 6, 0, (sockaddr *)&client, sizeof(client));
		}
		
		virtual ~connection();   // is deleted in state
		
		
		
		bool authorize(const char *name, const char *psswd) {
			passwd *pwd = getpwnam(name);
			
			if (pwd) {
				uid = pwd->pw_uid;
				gid = pwd->pw_gid;
			}		
		
			return pwd;
		}
		
		bool authorized() { return uid != -1; }
		
		void logout() { uid = -1; }
		
		
		server *serv();
		
	protected:

		virtual void removereq(request *co) {
			access();
			if (!requests.empty()) {
				requestset::iterator i = find(requests.begin(), requests.end(), co);
				if (i != requests.end() ) requests.erase(i);
			}
		}

		
		
		virtual void addreq(request *co) {
			requests.push_front(co);
		}
		
				
		
	public:
/*		void removereq(dword id) {
			requestset::iterator i = getrequest(id);
			if (i != requests.end() ) {
				request *r = *i;
				requests.erase(i);
				delete r;
			}
		}*/
		
	
		void answer(response *co, request *notify = 0);


};

inline request::request(connection *c, dword id) : conn(c), id(id), remove(false) { conn->addreq(this); }



class state {
		

	public:
		dword sessiontimeout;
		
	
	protected:
	friend class server;
	friend class connection;
	
		typedef std::vector<connection *> connset;
		connset connections;

	private:
		connset::iterator findconn(const sockaddr_in &addr) ;
		
		server *serv;
				
		
	public:
		state(server *serv) : serv(serv) { assert(serv); }

	
		connection *getconn(const sockaddr_in &addr) {
			connset::iterator i = findconn(addr);
			//if (i == 0 || !*i)
			//	throw server_error("No such connection");
			return i != 0 ? *i : 0;
		}


	public:
		void removeconn(const sockaddr_in &addr);
		void removeconn(connection *c);
		connection *addconn(const sockaddr_in &addr);

		connection *addconn(connection *co);
		
		connection *addconn();

};


inline int connection::idx() {
	return find(st->connections.begin(), st->connections.end(), this) - st->connections.begin();
}

inline server *connection::serv() { return st->serv; }




class packet;

class server {
	protected:
	friend class idpacket;
	friend class packet;
		stm::udpsrvsocketbstr sock;
		word port;

		static const dword bufsize = 1024;

		char buf[bufsize];

		sockaddr_in who;
		
		
		
//		typedef std::vector<pollfd> pollset;
//		pollset sockets;
		
		connection *currentconn;
		dword currentreqid;
		
		stm::filebinstream stderrbstr;
		
		typedef list<request *> subsrequests;
		
		//subsrequests notifysignal;
		subsrequests notifywrite;
		
		struct responseaddr {
			private:
				response *r;
			public:
				sockaddr_in who;
				request *notify;
				
				responseaddr(response *r, const sockaddr_in &who, request *notify) : r(r), who(who), notify(notify) {}
				operator response*() const { return r; }	
				response *operator->() const { return r; }
		};
				
			
		
		
		typedef list<responseaddr> responseaddrset;
		
		responseaddrset responsebuf;
		
		void passsignal();
		void passwrite();
		bool sendresbuf();
		
		
		void sendresponseto(response *r, const sockaddr_in &co);
		
		

	public:

		void serve();
		void answer(response *co);
		void answererror(response *co);
		void sendto(response *co, const sockaddr_in &who, request *notify = 0);
		
//		void subscribesignal(request *co);
		void subscribewrite(request *co) {
			notifywrite.push_back(co);				
		}
		
		/* is there any point in non-blocking writing ?
		   the longest operation we perform is reading a file asynchronously.
		   the only advantage would be taking and executing requests for reading file to buffer during waiting for sending.
		
		   The question is whether we should buffer the usual responses (i.e. except 64K block of file.
		   But I guess it can be implemented later only in server, with no changes to the packet interface.
		
		*/
		
		static const dword noreq = 0xFFFFFFFF;

		server(word port)
			: sock(port) , s(this) , stderrbstr(2), currentreqid(noreq), beforepoll(0),
			senderrorfunc(0), notifysignalfunc(0), packettype(0)
		{}

		~server();	
		
		void connadded(connection *c, int idx) {

/*			sockets.insert(sockets.begin() + idx + 1 , pollfd());
			sockets[idx].fd = c->sock.fd();
			sockets[idx].events = POLLIN;					*/
		}

		void connremoved(connection *c, int idx) {
//			sockets.erase(sockets.begin() + idx + 1);
		}


	public:
		string welcome;

		void regpacket(packet *co) {
			packets.push_front(co);
		}
		void deregpacket(packet *co) {
			if (!packets.empty())
				packets.remove(co);
		}

		void (*senderrorfunc)(server *s, int id, const char *co);
		void (*notifysignalfunc)();
		void (*beforepoll)();
		dword (*packettype)(const char *buf, dword count);

		packet *findpacket(dword type) const ;


	protected:
		typedef list<packet *> packetset;
		
		// packet really means "packet type",  not a "packet instance"
		
		packetset packets;
		state s;

	public:
		state *st() { return &s; }

		void senderror(int id, const char *co) {
			senderrorfunc(this, id, co);
		}

};


inline void connection::answer(response *co, request *notify) {
	st->serv->sendto(co, client, notify);
}



class packet {
	protected:
		connection * addconn() {
			return serv->st()->addconn(serv->who);		
		}
	
	public: //temp
		server *serv;


	public:
		dword type;

		packet(server *s, dword type) : serv(s) , type(type) { s->regpacket(this);  }
		virtual ~packet() { serv->deregpacket(this); }


		virtual void handle(connection *conn, const char *co, dword count) = 0;
};


class idpacket : public packet {
	protected:
		virtual dword id(const char *co, dword count) const = 0;
		
		request *getrequest(connection *conn, const char *co, dword count) { return conn->getrequest(id(co, count)); }
	
	public:
		idpacket(server *s, dword type) : packet(s, type) {}

		virtual void handle(connection *conn, const char *co, dword count) =0;
};


inline void idpacket::handle(connection *conn, const char *co, dword count)  {
	serv->currentreqid = id(co, count);
}

// packet can get connection session and request session




} // namespace fserver

//---------------------------------------------------------------------------
#endif
