#ifndef __STM_HPP
#define __STM_HPP

#include <exception>
#include <stdexcept>
#include <string.h>
#include <string>

#include <errno.h>

namespace stm {



class std_error : public std::exception {
	public:
		int code;
		
		std_error() : code(errno) { /*cerr << what();*/ }
		std_error(int code) : code(code) { /*cerr << what();*/ }
		std_error(const std_error &co) : std::exception(co), code(co.code) {}
		virtual const char * what() const  throw () {
			return strerror(code);
		}
		virtual const char *where()  const  { return "std_error"; }
};

class code_error : public std_error {
	public:
		code_error(int code) : std_error(code) {}
		code_error(const code_error &co) : std_error(co) {}
		
		virtual const char *where()  const  { return "code_error"; }
};


class wstd_error : public std_error {
	protected:
		std::string w;

	public:
		wstd_error() {}
		wstd_error(const wstd_error& co) : std_error(co), w(co.w) {}
		wstd_error(const char *where) : w(where) {}
		wstd_error(int code, const char *where) : w(where), std_error(code) {}
		wstd_error(int code) : std_error(code) {}

		virtual const char *where() const { return w.c_str() ? w.c_str() : std_error::where(); }
};

class wcode_error : public wstd_error {
	public:
		wcode_error(int code) : wstd_error(code) {}
		wcode_error(const wcode_error& co) : wstd_error(co) {}
		wcode_error(int code, const char *where) : wstd_error(code, where) {}				
};


class file_error : public wstd_error {
	protected:
		std::string f;

	public:
		file_error() {}
		file_error(const file_error &co)  : wstd_error(co), f(co.f) {}
		file_error(const char *where, const char *file) : wstd_error(where),  f(file) {}
		file_error(const char *file) : f(file) {}

		virtual const char *file()  const { return f.c_str() ? f.c_str() : "unknown file"; }
		virtual const char *where()  const { return w.c_str() ? w.c_str() : std_error::where(); }
};


inline dword checkerror(int ret, const char *where = "") {
	if (ret == -1)  {
		cout << "checkerror " << ret << endl;
		throw wstd_error(where);
	} else
		return ret;
}




} // namespace stm

#endif
