#ifndef __MAR_THREADS_HPP
#define __MAR_THREADS_HPP

#define __OLD_NNLIB_STRING

#include <exception>
#include <exception.hpp>

#include <pthread.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <functions.hpp>
#include <signal.h>
#include <sys/time.h>

#include <iostream>
#include <stm/stm.hpp>

namespace stm {


class thread {
	public:
/*		class toomany : public std::runtime_error {
			public:
				virtual const char *what() const throw () { return "Not enough system resources to create a process for a new thread"; }
		};

		class wrongjointhread : public std::runtime_error {
			public:
				virtual const char *what() const throw () { return "Thread was detached or another thread already is wating for it"; }
		};*/
	protected:
		pthread_t id;

		static void checkerror(int ret, const char *where) {
			if (ret) {
//				cerr << where << " : " << strerror(ret) << endl;
				throw wcode_error(ret, where);
			}
		}


		static void *pthreadroutine(void *arg) {
//			cout << "after fork child" << endl;
			return ((thread *)arg)->execute();
		}

		virtual void create() {
//			cout << "before fork" << endl;

			checkerror(pthread_create(&id, 0, pthreadroutine, this), "pthread_create");
//			cout << "after fork" << endl;
		}

		virtual void *execute() =0;
		void checkcancel() { pthread_testcancel(); }
		void exit(void *retval) { pthread_exit(retval); }
	public:
		thread() {}

		void resume() { create(); }
		void stop() { /* send a signal SIGSTOP, but it's not compatible with Unixes (at least Kylix guys says so) */}
		void terminate() { checkerror(pthread_cancel(id), "pthread_exit"); }
		pthread_t handle() { return id; }

		void *waitfor() {
			void *out;
			checkerror(pthread_join(id, &out), "pthread_join");
			return out;
		}

		void detach() { checkerror(pthread_detach(id), "pthread_detach"); }
};

// autodestruct thread
class adthread : public thread {
	public:
		static void *pthreadroutine(void *arg) {
			void *out = ((adthread *)arg)->execute();
			delete ((thread *)arg);
			return out;
		}

		virtual void create() {
			execute();
			checkerror(pthread_create(&id, 0, pthreadroutine, this), "pthread_create");
		}


		virtual ~adthread() {}

};

class cancelthread : public thread {
	protected:
		bool cancel;

		void checkcancel() { if (cancel) exit(PTHREAD_CANCELED); }
	public:
		cancelthread() : cancel(false) {}
		void terminate() { cancel = true; }

};









} // namespace stm


#endif

